/*
 FILE:PhysicsFS.h
urkle 	on:Tue Apr 30 13:34:44 2002

*/

#include	"PhysicsFS.h"

/*** PhysicsList ***/
// all inline

/*** PhysicsFS ***/
PhysicsFS::PhysicsFS()
{
	PHYSFS_Version compiled;
	PHYSFS_Version linked;
	
	PHYSFS_VERSION(&compiled);
	PHYSFS_getLinkedVersion(&linked);
	cout << "We compiled against PhysFS version " << 
					(int)compiled.major << "." << (int)compiled.minor << "." << (int)compiled.patch << "\n";
	cout << "But we linked against PhysFS version " <<
					(int)linked.major << "." << (int)linked.minor << "." << (int)linked.patch << "\n";
}

PhysicsFS::~PhysicsFS()
{
	this->Deinit();
}

const char *PhysicsFS::error()
{
	return PHYSFS_getLastError();
}

const char *PhysicsFS::dirSeperator()
{
	return PHYSFS_getDirSeparator();
}

bool PhysicsFS::Init(const char *basedir)
{
	return PHYSFS_init(basedir);
}

bool PhysicsFS::Deinit()
{
	return PHYSFS_deinit();
}

PhysicsList *PhysicsFS::cdromdirs()
{
	return new PhysicsList_CDROMS(PHYSFS_getCdRomDirs());
}

PhysicsList *PhysicsFS::archiveinfo()
{
	return new PhysicsList_ArchiveInfo(PHYSFS_supportedArchiveTypes());
}

PhysicsList *PhysicsFS::searchpath()
{
	return new PhysicsList_SearchPath(PHYSFS_getSearchPath());
}

bool PhysicsFS::symlinks()
{
	return _symlinks;
}

void PhysicsFS::symlinks(bool enabled)
{
	_symlinks = enabled;
	PHYSFS_permitSymbolicLinks(enabled);
}

const char *PhysicsFS::basedir()
{
	return PHYSFS_getBaseDir();
}

const char *PhysicsFS::userdir()
{
	return PHYSFS_getUserDir();
}

bool PhysicsFS::writedir(const char *dpath)
{
	return PHYSFS_setWriteDir(dpath);
}

const char *PhysicsFS::writedir()
{
	return PHYSFS_getWriteDir();
}

bool PhysicsFS::AddPath(const char *dpath, bool append)
{
	return PHYSFS_addToSearchPath(dpath,append);
}

bool PhysicsFS::DelPath(const char *dpath)
{
	return PHYSFS_removeFromSearchPath(dpath);
}

bool PhysicsFS::saneconfig(const char *organization,const char *appname, const char *archive_ext, bool include_cdroms, bool archives_first)
{
	strcpy(_organization,organization);
	strcpy(_appname,appname);
	return PHYSFS_setSaneConfig(organization,appname,archive_ext,(include_cdroms?1:0),(archives_first?1:0));
}

// file utility functions
bool PhysicsFS::mkdir(const char *ipath)
{
	return PHYSFS_mkdir(ipath);
}

bool PhysicsFS::rm(const char *ipath)
{
	return PHYSFS_delete(ipath);
}

bool PhysicsFS::exists(const char *ipath)
{
	return PHYSFS_exists(ipath);
}

bool PhysicsFS::isdir(const char *ipath)
{
	return PHYSFS_isDirectory(ipath);
}

bool PhysicsFS::issymlink(const char *ipath)
{
	return PHYSFS_isSymbolicLink(ipath);
}

PhysicsList *PhysicsFS::ls(const char *ipath)
{
	return new PhysicsList_ls(PHYSFS_enumerateFiles(ipath));
}

const char *PhysicsFS::realdir(const char*ipath)
{
	return PHYSFS_getRealDir(ipath);
}

/**** PhysicsFileBuf ****/
#define BUFSIZ 8192
PhysicsFileBuf::PhysicsFileBuf() : streambuf(),_file(NULL)
{
	char_type* gbuf = new char_type[BUFSIZ];
	//char_type* pbuf = new char_type[BUFSIZ];
	setg(gbuf,gbuf + BUFSIZ,gbuf + BUFSIZ);
	setp(gbuf,gbuf + BUFSIZ);
}

PhysicsFileBuf::~PhysicsFileBuf()
{
	if (this->is_open()) {
		this->close();
	}
	delete [] pbase();
	//delete [] eback();
}


PhysicsFileBuf *PhysicsFileBuf::open(const char *ipath,ios::openmode mode)
{
	_mode = mode;
	if (is_open())
		return NULL;

	if (_mode & ios::app) {
	//	cout << "Append\n";
		_file = PHYSFS_openAppend(ipath);
		_mode |= ios::out;
	} else if (_mode & ios::in) {
	//	cout << "read\n";
		_file = PHYSFS_openRead(ipath);
	} else if (_mode & ios::out) {
//			cout << "write\n";
		_file = PHYSFS_openWrite(ipath);
	} else {
		return NULL; //unsuported open mode
	}
	if (_file==NULL) {
		cout << "NULL" << PHYSFS_getLastError();
		return NULL;
	}
	return this;
}

PhysicsFileBuf *PhysicsFileBuf::close()
{
	//cout << "close\n";
	if (!this->is_open())
		return NULL;
	sync();	//make sure all is written;
	if (PHYSFS_close(_file)==0)
		return NULL;

	_file=NULL;
	return this;
}

bool PhysicsFileBuf::is_open()
{
	return (NULL!=_file);
}

/**** PhysicsFileBuf Protected members ****/
PhysicsFileBuf::pos_type PhysicsFileBuf::seekoff(off_type off, seekdir way, ios::openmode which=ios::in|ios::out)
{
	int pos=eof;
	switch (way) {
	case ios::beg:
		if (PHYSFS_seek(_file,off)!=0) {
			pos = off;
		}
		break;
	case ios::cur:
		pos = PHYSFS_tell(_file) - (egptr()-gptr());
		if (off==0)
			return pos;
		if (off!=0) {
			pos+=off;
			if (PHYSFS_seek(_file,pos)!=0) {
				setg(eback(),eback()+BUFSIZ,eback()+BUFSIZ);
			} else {
				pos=eof;
			}
		}
		break;
	case ios::end:
		pos = PHYSFS_fileLength(_file);
		pos-=off;
		if (PHYSFS_seek(_file,pos)!=0) {
			setg(eback(),eback()+BUFSIZ,eback()+BUFSIZ);
		} else {
			pos=eof;
		}
		break;
	}
	if (pos==eof)
		return eof;
	if (which & ios::in) {
		//cout <<"buff - in\n";
		setg(eback(),eback()+BUFSIZ,eback()+BUFSIZ);
	}
	if (which & ios::out) { //This never happens???
		//cout <<"buff - out\n";
		setp(pbase(),epptr());
	}
	return pos;
}

PhysicsFileBuf::int_type PhysicsFileBuf::overflow(int_type c)
{
	if (pbase()==0)
		return eof;
	if (c==eof)
		return sync();
	if (pptr()==epptr())
		sync();
	*pptr()=(char_type)c;
	pbump(1);
	return c;
}

PhysicsFileBuf::int_type PhysicsFileBuf::underflow()
{
	if(gptr() == 0)
		return eof;
	if (gptr() < egptr())
		return (unsigned char)*gptr();

	int rlen;
	rlen = PHYSFS_read(_file,eback(),1,egptr()-eback());
	if (rlen<=0)
		return eof;
	setg(eback(),eback(),eback()+rlen);
	return (unsigned char)*gptr();
}

int PhysicsFileBuf::uflow()
{
	int_type ret = underflow();
	if (ret == eof)
		return eof;
	gbump(1);
	return ret;
}

int PhysicsFileBuf::sync()
{
	if (pptr() && pbase() < pptr() && pptr() <= epptr()) {
		streamsize n = pptr() - pbase();
		int len = PHYSFS_write(_file,pbase(), 1,n);
		if (len<0)
			cerr << "Error writing to file\n";
		int wlen = ( pptr() - pbase() ) - len;
		if (wlen!=0) {
			//reposition unwritten chars
			char* pto = pbase();
			char* pfrom = pbase() + wlen;
			int len = pptr() - pbase() - wlen;
			while (pfrom < pptr())
				*pto++ = *pfrom++;
			setp(pbase(),epptr());
			pbump(len);
			cerr << "Error in Sync, cannot write "<<wlen<<" bytes.";
		}
		setp(pbase(),(char_type*)epptr());
	}
	return 0;
}

streamsize PhysicsFileBuf::xsgetn(char_type *s, streamsize n)
{
	int rval = showmanyc();
	if (rval >= n) {
		memcpy(s,gptr(),n * sizeof(char_type));
		gbump(n);
		return n;
	}
	memcpy(s, gptr(),rval * sizeof(char_type));
	gbump(rval);
	if (underflow() != eof)
		return rval + xsgetn(s + rval, n -rval);
	return rval;
}

streamsize PhysicsFileBuf::xsputn(char_type *s, streamsize n)
{
	int wval = epptr()-pptr();
	if (n <= wval) {
		memcpy(pptr(),s,n*sizeof(char_type));
		pbump(n);
		return n;
	}
	memcpy(pptr(),s,wval*sizeof(char_type));
	pbump(wval);

	if (overflow()!=eof)
		return wval + xsputn(s + wval, n-wval);
	return wval;
}

int PhysicsFileBuf::showmanyc() const
{
	if (gptr() && gptr() < egptr())
		return egptr() - gptr();
	return 0;
}

/**** PhysicsStream ****/
PhysicsStream::PhysicsStream() : iostream(),_PhysBuf(NULL)
{
	close();
}

PhysicsStream::PhysicsStream(const char *ipath, ios::openmode mode) : iostream(),_PhysBuf(NULL)
{
	open(ipath,mode);
}

PhysicsStream::~PhysicsStream()
{
	close();
	delete _PhysBuf;
}

void PhysicsStream::open(const char *ipath, ios::openmode mode)
{
	if (! _PhysBuf ) {
		this->rdbuf(_PhysBuf=new PhysicsFileBuf);
	}
	if (0==_PhysBuf->open(ipath,mode)) {
		setstate(badbit);
	}
}

void PhysicsStream::close()
{
	if (_PhysBuf) {
		_PhysBuf->close();
	}
}
