/*
 FILE:PhysicsFS.h
urkle 	on:Tue Apr 30 13:34:44 2002

*/

#ifndef	H_include_PhysicsFS
#define	H_include_PhysicsFS
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <physfs.h>
#include <iostream>

class PhysicsList {
protected:
	void **_listptr;
	void **_offset;
public:
	PhysicsList(void **listptr) : _listptr(listptr),_offset(listptr) {};
	virtual ~PhysicsList() { if (_listptr) PHYSFS_freeList(_listptr); };
	inline PhysicsList *operator++(int) {
		_offset++;
		return this;
	};
	inline PhysicsList *operator--(int) {
		_offset--;
		if (_offset<_listptr)
			_offset=_listptr;
		return this;
	};
	inline const char *val() {
		return (const char *)*_offset;
	}
};

class PhysicsList_ArchiveInfo : public PhysicsList {
public:
	PhysicsList_ArchiveInfo(const PHYSFS_ArchiveInfo **listptr) : PhysicsList((void **)listptr) {};
	virtual ~PhysicsList_ArchiveInfo() {_listptr=NULL;};

	inline const char *extension() {
			return ((PHYSFS_ArchiveInfo *)*_offset)->extension;
	};
	inline const char *description() {
			return ((PHYSFS_ArchiveInfo *)*_offset)->description;
	};
	inline const char *author() {
			return ((PHYSFS_ArchiveInfo *)*_offset)->author;
	};
	inline const char *url() {
			return ((PHYSFS_ArchiveInfo *)*_offset)->url;
	};
};

class PhysicsList_CDROMS : public PhysicsList {
public:
	PhysicsList_CDROMS(char **listptr) : PhysicsList((void **)listptr) {};
	virtual ~PhysicsList_CDROMS() {};
};

class PhysicsList_SearchPath : public PhysicsList {
public:
	PhysicsList_SearchPath(char **listptr) : PhysicsList((void **)listptr) {};
	virtual ~PhysicsList_SearchPath() {};
};

class PhysicsList_ls : public PhysicsList {
public:
	PhysicsList_ls(char **listptr) : PhysicsList((void **)listptr) {};
	virtual ~PhysicsList_ls() {};
	inline bool isdir() {
		return PHYSFS_isDirectory(val());
	};
	inline bool issymlink() {
		return PHYSFS_isSymbolicLink(val());
	};
};

class PhysicsFS {
private:
	bool _symlinks;
	char _organization[64];
	char _appname[64];
public:
	PhysicsFS();
	~PhysicsFS();

	const char *error();
	const char *dirSeperator();
// Setup and configuration	
	bool Init(const char *basedir); //basedir is a dpath
	bool Deinit();
	PhysicsList *archiveinfo();
	PhysicsList *cdromdirs();
	PhysicsList *searchpath();
	void symlinks(bool enabled);
	bool symlinks();
	const char *basedir(); //returns dpath
	const char *userdir(); //returns dpath
	const char *writedir(); //returns dpath
	bool writedir(const char *dpath);
	bool AddPath(const char *dpath, bool append=true);
	bool DelPath(const char *dpath);
	bool saneconfig(const char *organization,const char *appname, const char *archive_ext, bool include_cdroms, bool archives_first);
// File utility functions
	bool mkdir(const char *ipath);
	bool rm(const char *ipath);
	bool exists(const char *ipath);
	bool isdir(const char *ipath);
	bool issymlink(const char *ipath);
	PhysicsList *ls(const char *ipath);

	/* ls(const char *dir);
		list a DIR class???*/
//mapping function
	const char *realdir(const char*ipath); //returns dpath
};


class PhysicsFileBuf : public streambuf {
	PHYSFS_file *_file;
	ios::openmode _mode;
public:
	typedef char char_type;
	typedef streampos pos_type;
	typedef streamoff off_type;
	typedef int int_type;
	typedef ios::seek_dir seekdir;
	static const int_type eof = EOF;

	PhysicsFileBuf();
	~PhysicsFileBuf();
	PhysicsFileBuf *open (const char *ipath,ios::openmode mode=ios::in);
	PhysicsFileBuf *close();
	bool is_open();
protected:
	virtual pos_type seekoff(off_type off, seekdir way, ios::openmode which=ios::in|ios::out);
	virtual int sync();
	virtual int showmanyc() const;
	virtual streamsize xsgetn(char_type *s, streamsize n);
	virtual int_type underflow();
	virtual int_type uflow();
	virtual streamsize xsputn(char_type *s, streamsize n);	
	virtual int_type overflow(int_type c = eof);
};

class PhysicsStream : public iostream {
private:
	PhysicsFileBuf *_PhysBuf;
public:
	PhysicsStream(const char *ipath, ios::openmode mode=ios::in);
	PhysicsStream();
	~PhysicsStream();

	void open(const char *ipath, ios::openmode mode=ios::in);
	void close();
	
};
#endif	/*H_include_PhysicsFS*/


